/*
    Copyright Rene Rivera 2019

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef B2_EXAMPLE_PCH_MULTI_EXTRA_META_HPP
#define B2_EXAMPLE_PCH_MULTI_EXTRA_META_HPP

#include <type_traits>
#include <functional>
#include <typeindex>
#include <tuple>

#endif
