//  (C) Copyright Matt Borland 2021 - 2022.
//  Use, modification and distribution are subject to the
//  Boost Software License, Version 1.0. (See accompanying file
//  LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//

#ifndef BOOST_MATH_CCMATH_HPP
#define BOOST_MATH_CCMATH_HPP

#include <boost/math/ccmath/sqrt.hpp>
#include <boost/math/ccmath/isinf.hpp>
#include <boost/math/ccmath/isnan.hpp>
#include <boost/math/ccmath/abs.hpp>
#include <boost/math/ccmath/fabs.hpp>
#include <boost/math/ccmath/isfinite.hpp>
#include <boost/math/ccmath/isnormal.hpp>
#include <boost/math/ccmath/fpclassify.hpp>
#include <boost/math/ccmath/frexp.hpp>
#include <boost/math/ccmath/div.hpp>
#include <boost/math/ccmath/logb.hpp>
#include <boost/math/ccmath/ilogb.hpp>
#include <boost/math/ccmath/scalbn.hpp>
#include <boost/math/ccmath/scalbln.hpp>
#include <boost/math/ccmath/floor.hpp>
#include <boost/math/ccmath/ceil.hpp>
#include <boost/math/ccmath/trunc.hpp>
#include <boost/math/ccmath/modf.hpp>
#include <boost/math/ccmath/round.hpp>
#include <boost/math/ccmath/fmod.hpp>
#include <boost/math/ccmath/remainder.hpp>
#include <boost/math/ccmath/copysign.hpp>
#include <boost/math/ccmath/hypot.hpp>
#include <boost/math/ccmath/fdim.hpp>
#include <boost/math/ccmath/fmax.hpp>
#include <boost/math/ccmath/fmin.hpp>
#include <boost/math/ccmath/isgreater.hpp>
#include <boost/math/ccmath/isgreaterequal.hpp>
#include <boost/math/ccmath/isless.hpp>
#include <boost/math/ccmath/islessequal.hpp>
#include <boost/math/ccmath/isunordered.hpp>
#include <boost/math/ccmath/fma.hpp>

#endif // BOOST_MATH_CCMATH_HPP
